@@ -13,3 +13,6 @@ pub mod source;
 pub mod unalias;
 pub mod vox;
 pub mod ulimit;
+pub mod minfd;
+pub mod set;
+pub mod utils;
